@@ -1,15 +1,15 @@
-import json
-
+import orjson
 from requests import Response
 
 
-def _is_resp_ok(resp: Response) -> bool:
+def is_resp_ok(resp: Response) -> bool:
     if not resp.ok:
         return False
     if not resp.content:
         return False
     try:
-        d = resp.json()
-    except json.JSONDecodeError:
+        d = orjson.loads(resp.text)
+    except orjson.JSONDecodeError:
         return False
     return d['status'] == 'ok'
+
